/*
*
* Hangman Solver
*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 

This program is used to generate a solution path for all words in a dictionary. As output it generates a sequence of guesses for a word that guarantee a solution. I wrote this to find out what are the hardest words to guess. The surprising result is that long words (7-letter or more) are always guessed within 10 tries. It is the short ones (3-6 letters) that are difficult.

*/

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifndef WIN32
	#define _cdecl
	#undef PERF
#else
//	#define PERF
#endif

#ifdef PERF
#include <windows.h>
typedef __int64 int64;

class cHiresStopWatch
{
private:
	int64 iStart,iEnd;
	
	static int64 freq;

public:
	static bool Init(void)
	{
		return (0==::QueryPerformanceFrequency((LARGE_INTEGER*)&cHiresStopWatch::freq));
	}

public:
	cHiresStopWatch()
		:iStart(0)
		,iEnd(0)
	{
		this->Start();
	}

	void Start()
	{
		::QueryPerformanceCounter((LARGE_INTEGER*)&this->iStart);
	}

	void Stop()
	{
		::QueryPerformanceCounter((LARGE_INTEGER*)&this->iEnd);
	}

	int64 Freq() const
	{
		return freq;
	}

	int64 DeltaTicks() const{return (this->iEnd-this->iStart);}
	double DeltaSec() const {return (double)this->DeltaTicks()/(double)this->Freq();}
	double DeltaMilli() const {return 1000.0*(double)this->DeltaTicks()/(double)this->Freq();}
	double DeltaMicro() const {return 1000000.0*(double)this->DeltaTicks()/(double)this->Freq();}
};
int64  cHiresStopWatch::freq;
#endif

const char *dictname="web2";

char *rawdict=0;

bool matchsubstr(const char *szP, const char *szT)
{
	int i=0;
	char cp;
	while ((cp=szP[i]) & szT[i])
	{
		if(cp != '?')
		{
			if (cp != szT[i])
				return false;
		}

		++i;
	}

	return !cp && !szT[i];
}

class Dictionary
{
private:
	void _alloc(long wordcount)
	{
		this->words = (const char **) malloc(sizeof(const char*) * wordcount);
		if (this->words)
			this->count = wordcount;
	}

	void _zerohistogram(void)
	{
		for (int i=0; i<256; ++i)
			this->hist[i] = 0;
	}

	void _realloc(void)
	{
		this->words = (const char**) realloc(this->words, sizeof(const char*) * this->count);
	}

	void _copy(const Dictionary& d)
	{
		this->_alloc(d.count);
		if (this->words)
		{
			::memcpy(this->words, d.words, this->count* sizeof(const char*));
			::memcpy(this->hist, d.hist, sizeof(this->hist));
		}
	}

public:
	long count;
	const char **words;
	long hist[256];

	Dictionary()
		:words(0)
		,count(0)
	{
		this->_zerohistogram();
	}

	Dictionary(const Dictionary& d)
	{
		this->_copy(d);
	}

	void operator=(const Dictionary& d)
	{
		this->_copy(d);
	}

	~Dictionary()
	{
		if (this->words)
			free(this->words);	
	}

	void LoadRaw(const char* szRaw, long wordcount)
	{
		this->words = (const char **) malloc(sizeof(const char*) * wordcount);
		if (this->words)
		{
			this->count = wordcount;

			const char *sz = szRaw;
			const char **rgw = this->words;
			 
			for (long i=0; i<wordcount; ++i)
			{
				*rgw++ = sz;
				sz += ::strlen(sz)+1;
			}
		}
	}


	void Restrict(size_t len)
	{
		long iNew = 0;

		for (long i = 0; i<this->count; ++i)
		{
			if (len == ::strlen(this->words[i]))
			{
				this->words[iNew++] = this->words[i];
			}
		}
		this->count = iNew;
		this->_realloc();
	}

	void RestrictNot(char ch)
	{
		long iNew = 0;
		for (long i = 0; i<this->count; ++i)
		{
			if (!::strchr(this->words[i], ch))
			{
				this->words[iNew++] = this->words[i];
			}
		}
		this->count = iNew;
	}

	void Restrict(const char* sz)
	{
		long iNew = 0;
		for (long i = 0; i<this->count; ++i)
		{
			if (matchsubstr(sz, this->words[i]))
			{
				this->words[iNew++] = this->words[i];
			}
		}
		this->count = iNew;
	}

	void Dump(void)
	{
		for (long i = 0; i<this->count; ++i)
		{
			printf("%s\n", this->words[i]);
		}
	}


	void BuildHistogram(void)
	{
		this->_zerohistogram();
		for (long i = 0; i<this->count; ++i)
		{
			const char *sz = this->words[i];
			while(*sz)
				++this->hist[*sz++];
		}
	}

	void RestrictHistogram(const char* sz)
	{
		while(*sz)
		{
			this->hist[*sz] = 0;
			++sz;
		}
	}

	void DumpHistogram(void)
	{
		for (int i=0; i<256; ++i)
		{
			if (this->hist[i])
				printf("%c:%d ", i, this->hist[i]);
		}
		printf("\n");
	}

	char Guess(long *pcount)
	{
		long counthigh = -1;
		long ihigh;

		for (int i=0; i<256; ++i)
		{
			if (this->hist[i] > counthigh)
			{
				counthigh = this->hist[i];
				ihigh = i;
			}
		}

		*pcount = counthigh;

		this->hist[ihigh] = 0;
		return (char) ihigh;
	}

};

Dictionary dcFull;

const int kMaxDc=18;
Dictionary rgDc[kMaxDc];

bool filelen(const char* fn, size_t * plen)
{
	struct stat buf;
	if (0 == stat(fn, &buf))
	{
		*plen = buf.st_size;
		return false;
	}

	return true;
}

bool loaddict(void)
{
	bool fErr;
	size_t len;
	
	fErr = filelen(::dictname, &len);
	if(!fErr)
	{
		::rawdict = (char*)malloc(len+1);
		if (::rawdict)
		{
			FILE * pf;
			
			pf = fopen(::dictname, "r");
			if (pf)
			{
				size_t l;
			
				l = fread(::rawdict, 1, len, pf);

				//fErr = (len != l);
				len = l;
				fclose(pf);
			}

			if (!fErr)
			{
				size_t i;
				
				long wordcount=0;
				for (i=0; i<len; ++i)
				{
					if (::rawdict[i] == '\n')
					{
						::rawdict[i] = 0;
						++wordcount;
					}
					else
					{
						::rawdict[i] = ::tolower(::rawdict[i]);
					}
				}
				::rawdict[i] = 0;
				::dcFull.LoadRaw(::rawdict, wordcount);
				::dcFull.BuildHistogram();

				for (int d=1; d < ::kMaxDc; ++d)
				{
					::rgDc[d] = ::dcFull;
					::rgDc[d].Restrict(d);
					::rgDc[d].BuildHistogram();
				}
			}
		}
	}


	return fErr;
}

bool ismatched(const char *sz)
{
	while (*sz)
	{
		if (*sz++ == '?')
			return false;
	}
	return true;
}

bool tryguess(char ch, char *szMatch, const char* szAnswer)
{
	bool fNew = false;
	char * szM = szMatch;
	const char *szA = szAnswer;

	while (*szM)
	{
		if (*szM == '?' && ch == *szA)
		{
			fNew = true;
			*szM = ch;
		}

		++szM;
		++szA;
	}

	return fNew;
}
#ifdef PERF
static double cc=0;
static double tt=0;
cHiresStopWatch clk;
#endif

int matchword(const char *szWord, char *szGuesses)
{
	char szMatch[256];
	strcpy(szMatch, szWord);
	char *sz = szMatch;
	while(*sz)
		*sz++ = '?';

	int wordlen = ::strlen(szWord);
	Dictionary d((wordlen < kMaxDc)?::rgDc[wordlen]: ::dcFull);

	int failedattempts = 0;
	while (!ismatched(szMatch) && d.count)
	{
		long count;
		char guess = d.Guess(&count);

		if (count <= 0)
		{
			printf("unknown word\n");
			goto done;
		}

		sprintf(&szGuesses[::strlen(szGuesses)], ", %s/%c", szMatch, guess);

#ifdef PERF
::clk.Start();
::cc += d.count;
#endif
		if (tryguess(guess, szMatch, szWord))
			d.Restrict(szMatch);
		else
		{
			d.RestrictNot(guess);
			++failedattempts;
		}
#ifdef PERF
::clk.Stop();
::tt += ::clk.DeltaMilli();
#endif
		d.BuildHistogram();
		d.RestrictHistogram(szMatch);
	}

done:

	return failedattempts;
}

void _cdecl main(int argc, char *argv[])
{
	char *szword="";

	int cc = 0;
	if (argc == 2)
	{
		if (::isalpha(argv[1][0]))
			szword = argv[1];
		else
			cc = atoi(argv[1]);
	}

	if (loaddict())
		fprintf(stderr, "Error opening dictionary file\n");
	else
	{
#ifdef PERF
		cHiresStopWatch::Init();
#endif
	
		char szGuesses[100*256];
		int failedattempts;

		if (*szword)
		{
			if (strchr(szword, '?'))
			{
				::dcFull.Restrict(szword);
				::dcFull.Dump();
			}
			else
			{
				szGuesses[0] = 0;
				failedattempts = matchword(szword, szGuesses);
				printf("%s, %d%s\n", szword, failedattempts, szGuesses);
			}
		}
		else
		{
			for (int i = 0; i< (cc ? cc: ::dcFull.count); ++i)
			{
				szGuesses[0] = 0;
				failedattempts = matchword(::dcFull.words[i], szGuesses);
				printf("%s, %d%s\n", ::dcFull.words[i], failedattempts, szGuesses);
			}
#ifdef PERF
	printf("%g, %g, %gs\n", ::cc*1000/::tt, ::cc, ::tt/1000);
#endif
		}
	}
}
